"""Grand Unified SNAP - Bringing together SNAP Connect + SNAP2MQTT under one API"""

import asyncore
import base64
import binascii
from datetime import timedelta
import functools
import logging
import os
import random

from tornado.ioloop import IOLoop, PeriodicCallback, TimeoutError
from tornado.concurrent import Future
from tornado.gen import coroutine, with_timeout, sleep, Return

# SNAP Connect
import apy
from snapconnect import snap
from snappy import SnapConstantsHelper as SnapConst
import snaplib

from snapconnect_futures import SnapConnectFutures

from snap_profile.profile import SCSerialParams, SCTcpParams

# MQTT Support
from snap2mqtt_client.client import BaseClient
from snap2mqtt_future_client.future_client import FutureClient

log = logging.getLogger(__name__)


class GusError(Exception):
    """Wrapper for Gus-specific exceptions"""
    pass


class NoResponseError(Exception):
    """Exception raised if there's no response"""
    pass


class Gus(object):
    """The Grand Unified SNAP interface"""

    def __new__(cls, profile, network, *args, **kwargs):
        """The Gus Factory Function"""
        log.debug("In a gus factory")
        if profile.kind in ('serial', 'tcp'):
            return object.__new__(SCFGus)  # To avoid recursing
        elif profile.kind == 'mqtt':
            return object.__new__(MQTTGus)  # To avoid recursing

        else:
            raise GusError("Unknown kind in profile %r" % profile)

    def __init__(self):
        self._rpc_callbacks = {}
        self._data_callbacks = []
        self._traceroute_callbacks = []

    def __repr__(self):
        return '%s(%s, bridge=%s)' % (self.__class__.__name__, self.address, self.bridge_address)

    @property
    def bridge_address(self):
        """This should be the address of the bridge node.

        The bridge node is the node that gets us from serial to RF.
        """
        raise NotImplementedError

    @property
    def address(self):
        """This should be the 'self' address of the SNAP Interface.

        The self address is the address other nodes should send
        packets to if we want to receive them.
        """
        raise NotImplementedError

    def start(self):
        """Start Gus"""
        raise NotImplementedError

    def stop(self):
        """Stop Gus"""
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def on_any_rpc(self, callback):
        """Register callback for any rpc"""
        cbs = self._rpc_callbacks.setdefault('*', [])

        if callback not in cbs:
            cbs.append(callback)

    def on_rpc(self, func_name, callback):
        """Register callback for a specific rpc"""
        func_cbs = self._rpc_callbacks.setdefault(func_name, [])

        if callback not in func_cbs:
            func_cbs.append(callback)

    def on_traceroute(self, callback):
        """Register callback for traceroutes"""
        if callback not in self._traceroute_callbacks:
            self._traceroute_callbacks.append(callback)

    def on_data(self, callback):
        """Register callback for data packets"""
        if callback not in self._data_callbacks:
            self._data_callbacks.append(callback)

    def _on_rpc_recv(self, src, func, *args):
        """Call registered RPC callbacks"""
        log.debug("Got an RPC: %s.%s%r", src, func, args)

        # Handle the function-specific callbacks
        for cb in self._rpc_callbacks.get(func, []):
            try:
                IOLoop.current().add_callback(cb, src, *args)
            except:
                log.exception("RPC callback")

        # And the wildcard callbacks
        for cb in self._rpc_callbacks.get('*', []):
            try:
                IOLoop.current().add_callback(cb, src, func, *args)
            except:
                log.exception("RPC Wildcard callback")

    def _on_data_recv(self, src, data):
        """Call registered Data callbacks"""
        for cb in self._data_callbacks:
            try:
                IOLoop.current().add_callback(cb, src, data)
            except:
                log.exception("Data callback")

    def _on_traceroute_recv(self, src, rtt, hops):
        """Call registered Traceroute callbacks"""
        for cb in self._traceroute_callbacks:
            try:
                IOLoop.current().add_callback(cb, src, rtt, hops)
            except:
                log.exception("Traceroute callback")

    def ucast_rpc(self, target, func, *args, **kwargs):
        """Unicast RPC"""
        raise NotImplementedError

    def mcast_rpc(self, func, *args, **kwargs):
        """Multicast RPC"""
        raise NotImplementedError

    def dmcast_rpc(self, targets, func, *args, **kwargs):
        """Directed Multicast RPC"""
        raise NotImplementedError

    def ucast_data(self, target, data, **kwargs):
        """Unicast Data"""
        raise NotImplementedError

    def mcast_data(self, data, **kwargs):
        """Multicast Data"""
        raise NotImplementedError

    def callback_ucast_rpc(self, target, func, *args, **kwargs):
        """Callback Unicast RPC"""
        raise NotImplementedError

    def callback_mcast_rpc(self, func, *args, **kwargs):
        """Callback Multicast RPC"""
        raise NotImplementedError

    def callback_dmcast_rpc(self, targets, func, *args, **kwargs):
        """Callback Directed Multicast RPC"""
        raise NotImplementedError

    def callback_ucast_data(self, target, data, **kwargs):
        """Callback Unicast Data"""
        raise NotImplementedError

    def callback_mcast_data(self, data, **kwargs):
        """Callback Multicast Data"""
        raise NotImplementedError

    def traceroute(self, target):
        """Traceroute"""
        raise NotImplementedError

    def run_sync(self, timeout, cmd, *args, **kwargs):
        """Run the command to completion"""
        cb = functools.partial(cmd, self, *args, **kwargs)
        try:
            return IOLoop.current().run_sync(cb, timeout=timeout)
        except TimeoutError:
            if timeout is None:
                log.debug("Timeout error for command that would've run forever, not an error.")
            else:
                log.error("Command timed out or aborted!")
                raise

    @staticmethod
    def from_existing_scf(scf_instance, bridge_addr=None):
        """Create a Gus instance from an existing SCF instance rather than creating a new one."""

        gus_from_existing_scf = object.__new__(SCFGus)
        gus_from_existing_scf.__init__(None, None, None, scf_instance=scf_instance, bridge_addr=bridge_addr)
        return gus_from_existing_scf


class SCFGus(Gus):
    """Wraps SCF in a Gus-like interface"""

    def __init__(self, profile, network, sc_license, *args, **kwargs):
        """Create a new Gus"""

        super(SCFGus, self).__init__()

        # Check to see if SCFGus is being built from an existing SCF Instance
        self.bridge_addr = kwargs.get('bridge_addr', None)
        self._scf = kwargs.get('scf_instance', None)
        if self._scf:
            self._sc = self._scf._snap
            self.from_existing_instance = True
        else:
            self.from_existing_instance = False

        # If using an existing SCF Instance, this setup will have already been done.
        if not self.from_existing_instance:
            self.profile = profile.as_sc_params()
            self.network = network.as_sc_params()
            self.license = sc_license

            if self.license:
                path = os.path.dirname(os.path.realpath(__file__))
                license_path = os.path.join(
                    path, os.path.expanduser(self.license))
                log.debug("SNAP Connect License: %s", license_path)
            else:
                license_path = None
                log.debug("SNAP Connect License: Licensed by Ethernet MAC")

            self.nvparam_path = os.path.expanduser("~/.snap/nvparams.dat")
            log.debug("SNAP Connect NVParam: %s", self.nvparam_path)

            # We want a clean slate
            try:
                os.remove(self.nvparam_path)
            except:
                # Might not exist yet.
                pass

            if self.profile.address:
                log.debug("SNAP Connect Address: %s", binascii.hexlify(self.profile.address))
            else:
                log.debug("SNAP Connect Address: By Ethernet MAC or First in License File")

            self._sc = snap.Snap(scheduler=apy.ioloop_scheduler.IOLoopScheduler.instance(),
                                 license_file=license_path,
                                 nvparams_file=self.nvparam_path,
                                 addr=self.profile.address if license_path else None,
                                 funcs={})

            log.debug("SNAP Connect Licensed MAC: %s", binascii.hexlify(self._sc.licensed_mac))

            self._scf = SnapConnectFutures(self._sc)

            self.connect_future = None
            self.disconnect_future = None

            self._sc.set_hook(snap.hooks.HOOK_SNAPCOM_OPENED, self._on_connect)
            self._sc.set_hook(snap.hooks.HOOK_SNAPCOM_CLOSED, self._on_disconnect)

            self._sc.save_nv_param(snap.NV_AES128_ENABLE_ID, self.network.encryption_type)
            self._sc.save_nv_param(snap.NV_AES128_KEY_ID, self.network.encryption_key)

            # Mesh Group Params, we're interested in all, and might forward all.
            self._sc.save_nv_param(snap.NV_GROUP_INTEREST_MASK_ID, 0xFFFF)
            self._sc.save_nv_param(snap.NV_GROUP_FORWARDING_MASK_ID, 0xFFFF)
            self._sc.save_nv_param(snap.NV_GROUP_SERIAL_FORWARDING_MASK_ID, 0xFFFF)

            # We're just talking to our bridge, don't delete routes
            self._sc.save_nv_param(snap.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID, 0)

            # We need to randomize our mesh routing packets
            initial_mesh_seq_num = random.randint(1, 255)
            self._sc.save_nv_param(snap.NV_MESH_SEQUENCE_NUMBER_ID, initial_mesh_seq_num)

            # Also have to reach into the bowels of SC because updating
            # the config doesn't actually update this value...
            snaplib.MeshCodec.MeshCodec.sequence_number = initial_mesh_seq_num
            log.debug("MESH SEQ NUMBER INITIALIZED: %d", initial_mesh_seq_num)

            # Make room for packet CRCs, and send out data CRCs,
            # but don't worry about them actually being there...
            # SNAP Connect is a more forgiving about this stuff...
            self._sc.save_nv_param(snap.NV_FEATURE_BITS_ID, 0x0500)
            snaplib.RpcCodec.appendCrc = True
            snaplib.RpcCodec.validateCrc = False

        # Hook into the IOLoop
        self._cb = PeriodicCallback(self._sc.poll_internals, 5)

        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_MULTICAST_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_MULTICAST_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DIRECTED_MULTICAST_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DIRECTED_MULTICAST_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)

        # Also register for Data Mode packets.
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_BCAST, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_BCAST | SnapConst.PID_ENCRYPTION_MASK, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_PTP, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_PTP | SnapConst.PID_ENCRYPTION_MASK, self._on_data_recv)

        self._sc.set_hook(snap.hooks.HOOK_TRACEROUTE, self._on_traceroute_recv)

        # Alias functions up
        self.ucast_rpc = self._scf.rpc

    def _on_connect(self, connection_info, snap_address):
        """TCP connection callback"""
        log.info("Connection established: %s:%d, remote address is %s",
                 connection_info[0], connection_info[1], binascii.hexlify(snap_address))

        if self.connect_future is None:
            log.error("Unexpected TCP open call from %s:%d, remote address is %s",
                      connection_info[0], connection_info[1], binascii.hexlify(snap_address))
            return

        self.connect_future.set_result(binascii.hexlify(snap_address))

    def _on_disconnect(self, connection_info, snap_address):
        """TCP disconnect callback"""
        log.info("Connection Closed: %s:%d, remote address is %s",
                 connection_info[0], connection_info[1], binascii.hexlify(snap_address) if snap_address else 'None')

        if self.disconnect_future is None:
            log.error("Unexpected TCP close call from %s:%d", connection_info[0], connection_info[1])
            return

        self.disconnect_future.set_result(True)

    @property
    def bridge_address(self):
        return self.bridge_addr

    @property
    def address(self):
        return binascii.hexlify(self._sc.licensed_mac[-3:])

    def _on_rpc_recv(self, descriptor):
        super(SCFGus, self)._on_rpc_recv(binascii.hexlify(descriptor.srcAddr), descriptor.funcName, *descriptor.args)

    def _on_data_recv(self, descriptor):
        super(SCFGus, self)._on_data_recv(binascii.hexlify(descriptor.srcAddr), descriptor.data)

    def _on_traceroute_recv(self, src, rtt, hops):
        super(SCFGus, self)._on_traceroute_recv(binascii.hexlify(src), rtt, [{'addr': binascii.hexlify(hop[0]),
                                                                              'lq': hop[1]} for hop in hops])

    def mcast_rpc(self, func, *args, **kwargs):
        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        return self._scf.mcast_rpc(func, *args, group=group, ttl=ttl)

    def dmcast_rpc(self, targets, func, *args, **kwargs):
        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)
        delay_factor = kwargs.get('delay_factor', self.network.default_delay)

        return self._scf.dmcast_rpc(targets, func, *args, group=group, ttl=ttl, delay=delay_factor)

    @coroutine
    def callback_ucast_rpc(self, addr, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)

        result = yield self._scf.callback_rpc(addr, funcname, args, timeout=timeout, retries=retries,
                                              result_type=result_type, callback_name=callback_name)

        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_mcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        group = kwargs.get("group", self.network.default_group)

        result = yield self._scf.callback_mcast_rpc(addrs, funcname, args, timeout=timeout, retries=retries,
                                                    result_type=result_type, callback_name=callback_name,
                                                    ttl=ttl, group=group)

        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_dmcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        dmcast_delay = kwargs.get("dmcast_delay", self.network.default_delay)
        node_limit = kwargs.get("node_limit", None)
        expect_seq = kwargs.get("expect_seq", False)
        chunk_results = kwargs.get("chunk_results", False)
        previous_chunk = kwargs.get("previous_chunk", None)
        group = kwargs.get("group", self.network.default_group)

        result = yield self._scf.callback_dmcast(addrs, funcname, args, timeout=timeout, retries=retries,
                                                 result_type=result_type, callback_name=callback_name,
                                                 dmcast_delay=dmcast_delay, node_limit=node_limit, ttl=ttl,
                                                 expect_seq=expect_seq, chunk_results=chunk_results,
                                                 previous_chunk=previous_chunk, group=group)
        if result is None:
            raise NoResponseError()

        raise Return(result)

    def ucast_data(self, target, data, **kwargs):
        future = Future()
        log.info("DATA: %s.%r", target, data)

        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        rpc_id = self._scf._snap.data_mode(binascii.unhexlify(target), data)

        if self._scf._immediately_resolve_rpcs:
            future.set_result(True)
            self._scf._message_sent()
        else:
            self._scf._rpc_futures[rpc_id] = future

        return future

    def mcast_data(self, data, **kwargs):

        group = kwargs.get('group', self.network.default_group) or self.network.default_group
        ttl = kwargs.get('ttl', self.network.default_ttl) or self.network.default_ttl

        future = Future()
        log.info("DATA: %s,%s.%r", group, ttl, data)

        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        rpc_id = self._scf._snap.mcast_data_mode(group, ttl, data)

        if self._scf._immediately_resolve_rpcs:
            future.set_result(True)
            self._scf._message_sent()
        else:
            self._scf._rpc_futures[rpc_id] = future

        return future

    def traceroute(self, target):
        self._scf._snap.traceroute(binascii.unhexlify(target))

    @coroutine
    def _init_for_tcp(self):
        """Initialize for SNAPconnect TCP connection"""

        if self.connect_future and self.connect_future.running():
            raise Exception("Already connecting")

        self.connect_future = Future()

        # SNAP Connect TCP requires asyncore be polled as well.
        self._asyncpoll = PeriodicCallback(asyncore.poll, 5).start()

        def client_auth(realm):
            """SNAPconnect TCP client authentication"""
            return (self.profile.username, self.profile.password)

        self._scf._snap.connect_tcp(self.profile.host,
                                    auth_info=client_auth,
                                    port=self.profile.port,
                                    secure=self.profile.ssl)

        bridge_addr = yield with_timeout(timedelta(seconds=10), self.connect_future)

        if bridge_addr is None:
            raise Exception("Unable to open connection!")

        self.bridge_addr = bridge_addr

    @coroutine
    def _init_for_serial(self):
        """Initialize for SNAPconnect serial connection"""
        log.info("Opening serial type: %r, port %r", self.profile.type, self.profile.port)

        bridge_addr = yield self._scf.open_serial(self.profile.type, self.profile.port)

        if bridge_addr is None:
            raise Exception("Unable to open serial bridge!")

        self.bridge_addr = bridge_addr

    def start(self):
        log.debug("In start")

        self._cb.start()

        log.debug("callback started")

        # Check to see if SFCGus was built off an existing SFC Instance
        if not self.from_existing_instance:
            if isinstance(self.profile, SCSerialParams):
                try:
                    return IOLoop.current().run_sync(self._init_for_serial, timeout=30)
                except TimeoutError:
                    log.error("Timeout or abort starting serial!")
            elif isinstance(self.profile, SCTcpParams):
                try:
                    return IOLoop.current().run_sync(self._init_for_tcp, timeout=30)
                except TimeoutError:
                    log.error("Timeout or abort starting TCP!")
            else:
                raise GusError("%s is not a valid profile" % self.profile)

    @coroutine
    def _close_for_serial(self):
        """Shut down SNAPconnect serial connection"""
        # Give SNAP Connect a little bit longer to get packets out
        yield sleep(0.5)

        yield self._scf.close_serial(self.profile.type, self.profile.port)

    @coroutine
    def _close_for_tcp(self):
        """Shut down SNAPconnect TCP connection"""
        # Give SNAP Connect a little bit longer to get packets out
        yield sleep(0.5)

        if self.disconnect_future and self.disconnect_future.running():
            raise Exception("Already disconnecting")

        self.disconnect_future = Future()

        self._scf._snap.disconnect_tcp(self.profile.host, port=self.profile.port)

        yield with_timeout(timedelta(seconds=10), self.disconnect_future)

    def stop(self):
        # Close things down
        # Check to see if SFCGus was built off an existing SFC Instance
        if not self.from_existing_instance:
            if isinstance(self.profile, SCSerialParams):
                try:
                    return IOLoop.current().run_sync(self._close_for_serial, timeout=30)
                except TimeoutError:
                    log.error("Timeout or abort closing serial!")
            elif isinstance(self.profile, SCTcpParams):
                try:
                    return IOLoop.current().run_sync(self._close_for_tcp, timeout=30)
                except TimeoutError:
                    log.error("Timeout or abort closing TCP!")
            else:
                raise GusError("%s is not a valid profile" % self.profile)

        self._cb.stop()

        # Clean up after yourself
        try:
            os.remove(self.nvparam_path)
        except:
            pass


class MQTTGus(Gus):
    """Wraps SNAP2MQTTFutures in a Gus-like interface"""

    def __init__(self, profile, network, *args, **kwargs):
        """Create a new Gus"""

        super(MQTTGus, self).__init__()

        log.debug("New MQTTGus with profile: %r", profile)

        self.addr = profile.address

        self.profile = profile
        self.network = network

        self._base_client = BaseClient(profile.address, mqtt_host=profile.host)
        self._client = FutureClient(self._base_client)

        # Alias commands up
        self.ucast_rpc = self._client.uRPC

    @property
    def bridge_address(self):
        return self._base_client.bridge_address

    @property
    def address(self):
        return self.addr

    @coroutine
    def _init_for_mqtt(self):
        """Connect the MQTT client"""

        self._base_client.start()

        connect_future = Future()
        address_future = Future()

        def on_connect(client, userdata, flags, rc):
            """MQTT connect callback"""
            log.info("Connected to MQTT with result code %s", rc)
            connect_future.set_result(True)

        def on_address_change(address):
            log.info("MQTT Bridge Address: %r", address)
            try:
                address_future.set_result(True)
            except:
                log.exception("Got multiple address changed notifications")

        self._base_client.on_connect(on_connect)
        self._base_client.on_address_change(on_address_change)

        self._base_client.subscribe('rpc/#', self._on_rpc_recv)
        self._base_client.subscribe('data/#', self._on_data_recv)
        self._base_client.subscribe('traceroute/#', self._on_traceroute_recv)

        yield with_timeout(timedelta(seconds=10), connect_future)
        yield with_timeout(timedelta(seconds=10), address_future)

    def start(self):
        try:
            return IOLoop.current().run_sync(self._init_for_mqtt, timeout=30)
        except TimeoutError:
            log.error("Timeout or abort starting MQTT!")

    def stop(self):
        """Stop the MQTT client"""
        self._base_client.stop()

    def _on_rpc_recv(self, pkt):
        super(MQTTGus, self)._on_rpc_recv(pkt['src'], pkt['func'], *pkt['args'])

    def _on_data_recv(self, pkt):
        super(MQTTGus, self)._on_data_recv(pkt['src'], pkt['data'])

    def _on_traceroute_recv(self, pkt):
        super(MQTTGus, self)._on_traceroute_recv(pkt['src'], pkt['rtt'], pkt['hops'])

    def mcast_rpc(self, funcname, *args, **kwargs):
        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        return self._client.mRPC(group, ttl, funcname, *args)

    def dmcast_rpc(self, addrs, funcname, *args, **kwargs):

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)
        delay = kwargs.get('delay', self.network.default_delay)

        return self._client.dmRPC(addrs, group, ttl, delay, funcname, *args)

    def ucast_data(self, addr, data, **kwargs):
        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        return self._client.uData(addr, data)

    def mcast_data(self, data, **kwargs):

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        return self._client.mData(group, ttl, data)

    @coroutine
    def callback_ucast_rpc(self, addr, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)

        result = yield self._client.callback_rpc(addr, funcname, args, timeout=timeout, retries=retries,
                                                 result_type=result_type, callback_name=callback_name)
        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_mcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        group = kwargs.get("group", self.network.default_group)
        ttl = kwargs.get("ttl", self.network.default_ttl)

        result = yield self._client.callback_mcast_rpc(addrs, funcname, args,
                                                       timeout=timeout, retries=retries,
                                                       result_type=result_type,
                                                       callback_name=callback_name,
                                                       group=group, ttl=ttl)
        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_dmcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", 3)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        group = kwargs.get("group", self.network.default_group)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        dmcast_delay = kwargs.get("dmcast_delay", self.network.default_delay)
        node_limit = kwargs.get("node_limit", None)
        chunk_results = kwargs.get("chunk_results", False)
        previous_chunk = kwargs.get("previous_chunk", None)

        result = yield self._client.callback_dmcast(addrs, funcname, args,
                                                    timeout=timeout, retries=retries,
                                                    result_type=result_type,
                                                    callback_name=callback_name,
                                                    dmcast_delay=dmcast_delay,
                                                    node_limit=node_limit, ttl=ttl,
                                                    chunk_results=chunk_results,
                                                    previous_chunk=previous_chunk,
                                                    group=group)

        if result is None:
            raise NoResponseError()

        raise Return(result)

    def traceroute(self, target):
        return self._client.traceroute(target)
