# Node-level operations/abstractions
import logging

import binascii
import bz2
from collections import OrderedDict, Counter, defaultdict
import importlib

from .flash import _get_file_path_for_version

# Helpers
import formatters
from util import first_or_none, ensure_path_contains_python_lib
from nvparam import get_nvparam

import os

from tornado.concurrent import Future
from tornado.gen import coroutine, sleep, Return

from snapconnect import snap

from snaplib import ScriptsManager

from snappy import SnapConstantsHelper

from .util import fix_up_target, CRYPTO_LOOKUP
from .gus import NoResponseError

log = logging.getLogger(__name__)


@coroutine
def query_value_with_data(gus, target, data, key, rpc, *args, **kwargs):
    formatter = kwargs.get('fmt', lambda x: x)

    try:
        val = yield gus.callback_ucast_rpc(target, rpc, *args)
    except NoResponseError:
        val = None

    if val is None:
        data['_errors'] += 1

    val = first_or_none(val)

    if val is not None:
        data[key] = formatter(val)
    else:
        data[key] = None


@coroutine
def query_nvparams_with_data(gus, target, data, nvparams):
    """
    This helper function queries multiple nvparams. The

    Args:
        nvparams (dict): Dictionary of nvparam definitions; the key is the
            nvparam id and the value is the key under which the value will
            be stored in data

    Returns:
        None, though data will be populated.
    """

    for nvparam_id, data_key in nvparams.iteritems():
        try:
            log.debug("Getting NVParam for %r: %r", target, nvparam_id)
            val = yield get_nvparam(gus, target, nvparam_id)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
            data[data_key] = val
            return
        data[data_key] = val.values()[0]


@coroutine  # noqa
@fix_up_target
def device_info(gus, target):
    data = {}
    data['_errors'] = 0

    device_nvparams = OrderedDict([
        (snap.NV_MAC_ADDR_ID, 'device.address'),
        (snap.NV_DEVICE_NAME_ID, 'device.name'),
        (snap.NV_DEVICE_TYPE_ID, 'device.type'),
        (snap.NV_FEATURE_BITS_ID, 'device.feature-bits'),
        (snap.NV_SYS_PLATFORM_ID, 'device.platform'),
        (snap.NV_VENDOR_SETTINGS_ID, 'device.vendor-settings'),
    ])

    yield query_nvparams_with_data(gus, target, data, device_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def firmware_info(gus, target):
    data = {}
    data['_errors'] = 0

    val = yield get_vmstat(gus, target, 6)

    info = val['data'].split(',')

    if len(info) >= 8:
        core = info[0]
        encryption = CRYPTO_LOOKUP[int(info[3])]
        module_id = int(info[5])
        platform_str = info[7]
        try:
            firmware_category = SnapConstantsHelper.platforms_ids_to_strings[module_id]
        except KeyError:
            firmware_category = "Unknown"

        data['device.core'] = core
        data['firmware.category-id'] = module_id
        data['device.platform'] = platform_str
        data['firmware.version'] = "{} with {} encryption".format(core, encryption)
        data['firmware.category'] = firmware_category

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def mcast_info(gus, target):
    data = {}
    data['_errors'] = 0

    mcast_nvparams = OrderedDict([
        (SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID, 'mcast.processed-groups'),
        (SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID, 'mcast.forwarded-groups'),
        (SnapConstantsHelper.NV_GROUP_SERIAL_FORWARDING_MASK_ID, 'mcast.serial-forwarded-groups'),
        (SnapConstantsHelper.NV_CARRIER_SENSE_ID, 'mcast.carrier-sense'),
        (SnapConstantsHelper.NV_COLLISION_DETECT_ID, 'mcast.collision-detect'),
        (SnapConstantsHelper.NV_COLLISION_AVOIDANCE_ID, 'mcast.collision-avoidance'),
        (SnapConstantsHelper.NV_CS_CD_LEVEL_ID, 'mcast.cs-cd-threshold'),
        (SnapConstantsHelper.NV_CSMA_TIMESLOT_SETTINGS_ID, 'mcast.csma-settings'),
    ])
    yield query_nvparams_with_data(gus, target, data, mcast_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def mesh_info(gus, target):
    data = {}
    data['_errors'] = 0

    mesh_nvparams = OrderedDict([
        (SnapConstantsHelper.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID, 'mesh.routing.max-timeout'),
        (SnapConstantsHelper.NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID, 'mesh.routing.min-timeout'),
        (SnapConstantsHelper.NV_MESH_ROUTE_NEW_TIMEOUT_ID, 'mesh.routing.new-timeout'),
        (SnapConstantsHelper.NV_MESH_ROUTE_USED_TIMEOUT_ID, 'mesh.routing.used-timeout'),
        (SnapConstantsHelper.NV_MESH_ROUTE_DELETE_TIMEOUT_ID, 'mesh.routing.delete-timeout'),
        (SnapConstantsHelper.NV_MESH_RREQ_TRIES_ID, 'mesh.routing.rreq-retries'),
        (SnapConstantsHelper.NV_MESH_RREQ_WAIT_TIME_ID, 'mesh.routing.rreq-wait-timeout'),
        (SnapConstantsHelper.NV_MESH_INITIAL_HOPLIMIT_ID, 'mesh.routing.initial-hop-limit'),
        (SnapConstantsHelper.NV_MESH_MAX_HOPLIMIT_ID, 'mesh.routing.max-hop-limit'),
        (SnapConstantsHelper.NV_MESH_SEQUENCE_NUMBER_ID, 'mesh.seq-number'),
        (SnapConstantsHelper.NV_MESH_OVERRIDE_ID, 'mesh.override'),
        (SnapConstantsHelper.NV_MESH_PENALTY_LQ_THRESHOLD_ID, 'mesh.penalty-lq-threshold'),
        (SnapConstantsHelper.NV_MESH_REJECTION_LQ_THRESHOLD_ID, 'mesh.rejection-lq-threshold'),
    ])
    yield query_nvparams_with_data(gus, target, data, mesh_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def network_info(gus, target):
    data = {}
    data['_errors'] = 0

    radio_nvparams = OrderedDict([
        (snap.NV_CHANNEL_ID, 'net.channel'),
        (snap.NV_NETWORK_ID, 'net.network-id'),
        (SnapConstantsHelper.NV_SNAP_MAX_RETRIES_ID, 'net.radio-unicast-retries'),
        (SnapConstantsHelper.NV_SNAP_LQ_THRESHOLD_ID, 'net.radio-lq-threshold'),
        (SnapConstantsHelper.NV_RADIO_RATE_ID, 'net.default-radio-rate'),
    ])
    yield query_nvparams_with_data(gus, target, data, radio_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def script_info(gus, target):
    data = {}
    data['_errors'] = 0

    yield query_value_with_data(gus, target, data, 'device.script-name', 'imageName')
    yield query_nvparams_with_data(gus, target, data, {snap.NV_SNAPPY_CRC_ID: 'device.script-crc'})

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def security_info(gus, target):
    data = {}
    data['_errors'] = 0

    security_nvparams = OrderedDict([
        (SnapConstantsHelper.NV_AES128_ENABLE_ID, 'security.enable-encryption'),
        # (SnapConstantsHelper.NV_AES128_KEY_ID, 'encryption_key'),  # Explicitly not asking for this, if you can ask, you already know it.
        (SnapConstantsHelper.NV_LOCKDOWN_FLAGS_ID, 'security.lockdown-bitmask'),
    ])
    yield query_nvparams_with_data(gus, target, data, security_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def stats_info(gus, target):
    data = {}
    data['_errors'] = 0

    # Get all stat stats
    for stat in stat_lookup:
        if stat.startswith('stat'):
            result = yield get_stats(gus, target, stat)
            data.update(**result)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def summary_info(gus, target):
    data = {}
    data['_errors'] = 0

    # Live parameters
    yield query_value_with_data(gus, target, data, 'current.channel', 'getChannel')
    yield query_value_with_data(gus, target, data, 'current.network-id', 'getNetId', fmt=formatters.format_as_hex)

    device_data = yield device_info(gus, target)
    data.update(**device_data)

    script_data = yield script_info(gus, target)
    data.update(**script_data)

    firmware_data = yield firmware_info(gus, target)
    data.update(**firmware_data)

    security_data = yield security_info(gus, target)
    data.update(**security_data)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def uart_info(gus, target):
    data = {}
    data['_errors'] = 0

    uart_nvparams = OrderedDict([
        (SnapConstantsHelper.NV_DEFAULT_UART_ID, 'uart.default'),
        (SnapConstantsHelper.NV_UART_DM_TIMEOUT_ID, 'uart.buffering-timeout'),
        (SnapConstantsHelper.NV_UART_DM_THRESHOLD_ID, 'uart.buffering-threshold'),
        (SnapConstantsHelper.NV_UART_DM_INTERCHAR_ID, 'uart.intercharacter-timeout'),
        (SnapConstantsHelper.NV_UART0_BAUDRATE_ID, 'uart.uart0-default-baudrate'),
        (SnapConstantsHelper.NV_UART1_BAUDRATE_ID, 'uart.uart1-default-baudrate'),
    ])
    yield query_nvparams_with_data(gus, target, data, uart_nvparams)

    raise Return(data)


@coroutine  # noqa
@fix_up_target
def get_info(gus, target, verbosity=0):
    """Get node info from the target node.

    C.f. Portal's Node Info and Parameters pane.

    Args:
        gus (Gus): Grand Unified SNAP instance
        target (str, SNAP Address): Target Node
        verbosity (int): How deep do we go?

    Returns:
        A dict with the information

    """

    data = {}

    data['_errors'] = 0

    @coroutine
    def query_value(key, rpc, *args, **kwargs):
        formatter = kwargs.get('fmt', lambda x: x)

        try:
            val = yield gus.callback_ucast_rpc(target, rpc, *args)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1

        val = first_or_none(val)

        if val is not None:
            data[key] = formatter(val)
        else:
            data[key] = None

    @coroutine
    def query_vmstat(key, vmstat, *args, **kwargs):
        formatter = kwargs.get('fmt', lambda c, h, d: (c, h, d))
        args = [vmstat] + list(args)
        try:
            val = yield gus.callback_ucast_rpc(target, 'vmStat', *args, callback_name="tellVmStat")
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1
            data[key] = None

        if val is not None:
            (word, d) = val

            word = word & 0xFFFF

            code = word & 0xFF
            hibyte = word >> 8

            if code == vmstat:
                data[key] = formatter(code, hibyte, d)
            else:
                data[key] = None

    @coroutine
    def query_nvparams(nvparams):
        """
        This helper function queries multiple nvparams. The

        Args:
            nvparams (dict): Dictionary of nvparam definitions; the key is the
                nvparam id and the value is the key under which the value will
                be stored in data

        Returns:
            None, though data will be populated.
        """

        for nvparam_id, data_key in nvparams.iteritems():
            try:
                val = yield gus.callback_ucast_rpc(target, "loadNvParam", nvparam_id)
            except NoResponseError:
                val = None

            if val is None:
                data['_errors'] += 1
            data[data_key] = first_or_none(val)

    yield query_value('address', 'loadNvParam', snap.NV_MAC_ADDR_ID, fmt=binascii.hexlify)
    yield query_value('current_channel', 'getChannel')
    yield query_value('current_netid', 'getNetId', fmt=formatters.format_as_hex)
    yield query_value('nv_channel', 'loadNvParam', snap.NV_CHANNEL_ID)
    yield query_value('nv_net_id', 'loadNvParam', snap.NV_NETWORK_ID, fmt=formatters.format_as_hex)
    yield query_value('nv_feature_bits', 'loadNvParam', snap.NV_FEATURE_BITS_ID, fmt=formatters.format_as_hex)
    yield query_value('nv_vendor_settings', 'loadNvParam', snap.NV_VENDOR_SETTINGS_ID, fmt=formatters.format_as_hex)

    if verbosity > 0:
        yield query_value('nv_device_name', 'loadNvParam', snap.NV_DEVICE_NAME_ID)
        yield query_value('nv_device_type', 'loadNvParam', snap.NV_DEVICE_TYPE_ID)
        yield query_value('script_name', 'imageName')
        yield query_value('script_crc', 'loadNvParam', snap.NV_SNAPPY_CRC_ID, fmt=formatters.format_as_hex)
        yield query_vmstat('space_available', SnapConstantsHelper.VM_SPACE, fmt=lambda c, h, d: formatters.format_as_unsigned(d))
        yield query_value('platform', 'loadNvParam', snap.NV_SYS_PLATFORM_ID)

        def fmt_version(code, hibyte, data):
            version_info = data.split(',')
            core_version = version_info[0]
            encryption = CRYPTO_LOOKUP[int(version_info[3])]
            return "{} with {} encryption".format(core_version, encryption)

        yield query_vmstat('firmware_version', 6, fmt=fmt_version)

        # Getting firmware category is also a bit more complex...
        try:
            val = yield gus.callback_ucast_rpc(target, "getInfo", 3)
        except NoResponseError:
            val = None

        if val is None:
            data['_errors'] += 1

        firmware_category_id = first_or_none(val)

        try:
            data['firmware_category'] = SnapConstantsHelper.platforms_ids_to_strings[firmware_category_id]
        except KeyError:
            data['firmware_category'] = "Unknown"

    if verbosity > 1:
        # Get even more info: Radio, UART, Multicast
        radio_nvparams = {
            SnapConstantsHelper.NV_SNAP_MAX_RETRIES_ID: 'radio_unicast_retries',
            SnapConstantsHelper.NV_SNAP_LQ_THRESHOLD_ID: 'radio_lq_threshold',
        }
        yield query_nvparams(radio_nvparams)

        uart_nvparams = {
            SnapConstantsHelper.NV_DEFAULT_UART_ID: 'uart_default_uart',
            SnapConstantsHelper.NV_UART_DM_TIMEOUT_ID: 'uart_dm_timeout',
            SnapConstantsHelper.NV_UART_DM_THRESHOLD_ID: 'uart_dm_threshold',
            SnapConstantsHelper.NV_UART_DM_INTERCHAR_ID: 'uart_dn_interchar',
            SnapConstantsHelper.NV_UART0_BAUDRATE_ID: 'uart_uart0_baudrate',
            SnapConstantsHelper.NV_UART1_BAUDRATE_ID: 'uart_uart1_baudrate',
        }
        yield query_nvparams(uart_nvparams)

        mcast_nvparams = {
            SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID: 'mcast_group_interest_mask',
            SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID: 'mcast_group_forwarding_mask',
            SnapConstantsHelper.NV_GROUP_SERIAL_FORWARDING_MASK_ID: 'mcast_serial_forwarding_mask',
            SnapConstantsHelper.NV_CARRIER_SENSE_ID: 'mcast_carrier_sense',
            SnapConstantsHelper.NV_COLLISION_DETECT_ID: 'mcast_collision_detect',
            SnapConstantsHelper.NV_COLLISION_AVOIDANCE_ID: 'mcast_collision_avoidance',
            SnapConstantsHelper.NV_CS_CD_LEVEL_ID: 'mcast_cs_cd_level',
            SnapConstantsHelper.NV_CSMA_TIMESLOT_SETTINGS_ID: 'mcast_csma_timeslot_settings',
        }
        yield query_nvparams(mcast_nvparams)

        data['firmware_category_id'] = firmware_category_id

    if verbosity > 2:
        # Get all the info: Mesh
        mesh_nvparams = {
            SnapConstantsHelper.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID: 'mesh_route_max_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID: 'mesh_route_min_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_NEW_TIMEOUT_ID: 'mesh_route_new_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_USED_TIMEOUT_ID: 'mesh_route_used_timeout',
            SnapConstantsHelper.NV_MESH_ROUTE_DELETE_TIMEOUT_ID: 'mesh_route_delete_timeout',
            SnapConstantsHelper.NV_MESH_RREQ_TRIES_ID: 'mesh_rreq_tries',
            SnapConstantsHelper.NV_MESH_RREQ_WAIT_TIME_ID: 'mesh_wait_time',
            SnapConstantsHelper.NV_MESH_INITIAL_HOPLIMIT_ID: 'mesh_initial_hop_limit',
            SnapConstantsHelper.NV_MESH_MAX_HOPLIMIT_ID: 'mesh_max_hop_limit',
            SnapConstantsHelper.NV_MESH_SEQUENCE_NUMBER_ID: 'mesh_seq_number',
            SnapConstantsHelper.NV_MESH_OVERRIDE_ID: 'mesh_override',
            SnapConstantsHelper.NV_MESH_PENALTY_LQ_THRESHOLD_ID: 'mesh_penalty_lq_threshold',
            SnapConstantsHelper.NV_MESH_REJECTION_LQ_THRESHOLD_ID: 'mesh_rejection_lq_threshold',
        }
        yield query_nvparams(mesh_nvparams)

    raise Return(data)


@coroutine
@fix_up_target
def reboot(gus, target, delay=None):
    """Reboot the target node.

    Equivalent of the reboot(delay) command.

    Args:
        gus (Gus): Grand Unified SNAP instance
        target (str, SNAP Address): Target Node
        delay (Optional[int]): (Optional) ms to wait before rebooting

    Returns:
        A dict with the response from the reboot callback.

    """

    data = {}

    if delay is None:
        val = yield gus.callback_ucast_rpc(target, "reboot")
    else:
        val = yield gus.callback_ucast_rpc(target, "reboot", delay)
    data['reboot'] = first_or_none(val)

    raise Return(data)


@coroutine
@fix_up_target
def script_erase(gus, target):
    """OTA Script erase from the target node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        target (str, SNAP Address): Target Node

    Returns:
        A dict with the response

    """
    data = {}

    yield gus.callback_ucast_rpc(target, 'eraseImage')

    raise Return(data)


stat_lookup = OrderedDict([
    ('info.vendor', {'id': 0, 'rpc': 'getInfo', 'formatter': formatters.format_as_vendor}),
    ('info.primary-communications-interface', {'id': 1, 'rpc': 'getInfo', 'formatter': formatters.format_as_interface}),
    ('info.cpu', {'id': 2, 'rpc': 'getInfo', 'formatter': formatters.format_as_cpu}),
    ('info.platform', {'id': 3, 'rpc': 'getInfo', 'formatter': formatters.format_as_platform}),
    ('info.build', {'id': 4, 'rpc': 'getInfo', 'formatter': formatters.format_as_build}),

    ('info.version.major', {'id': 5, 'rpc': 'getInfo'}),
    ('info.version.minor', {'id': 6, 'rpc': 'getInfo'}),
    ('info.version.build', {'id': 7, 'rpc': 'getInfo'}),

    ('info.rpc.buffer', {'id': 9, 'rpc': 'getInfo'}),
    ('info.rpc.is-multicast', {'id': 10, 'rpc': 'getInfo'}),
    ('info.rpc.remaining-ttl', {'id': 11, 'rpc': 'getInfo'}),
    ('info.rpc.is-dmcast', {'id': 24, 'rpc': 'getInfo'}),

    ('info.mcast.groups', {'id': 27, 'rpc': 'getInfo'}),

    ('info.dmcast.address-index', {'id': 26, 'rpc': 'getInfo'}),
    ('info.dmcast.original-ttl', {'id': 28, 'rpc': 'getInfo'}),

    ('info.strings.tiny-left', {'id': 19, 'rpc': 'getInfo'}),
    ('info.strings.small-left', {'id': 12, 'rpc': 'getInfo'}),
    ('info.strings.medium-left', {'id': 13, 'rpc': 'getInfo'}),
    ('info.strings.large-left', {'id': 20, 'rpc': 'getInfo'}),

    ('info.script.is-first-execution', {'id': 21, 'rpc': 'getInfo'}),
    ('info.script.base-address', {'id': 22, 'rpc': 'getInfo'}),
    ('info.script.base-bank', {'id': 23, 'rpc': 'getInfo'}),

    ('info.encryption-available', {'id': 8, 'rpc': 'getInfo', 'formatter': formatters.format_as_encryption_status}),
    ('info.route-table-size', {'id': 14, 'rpc': 'getInfo'}),
    ('info.route-table-entries', {'id': 15, 'rpc': 'getInfo'}),
    ('info.bank-free-space', {'id': 16, 'rpc': 'getInfo'}),
    ('info.stdin-hook-trigger', {'id': 18, 'rpc': 'getInfo'}),
    ('info.read-and-reset-delay-factor', {'id': 25, 'rpc': 'getInfo'}),

    ('stat.ds-null.tx-buffers', {'id': 0, 'rpc': 'getStat'}),
    ('stat.ds-uart0.rx-buffers', {'id': 1, 'rpc': 'getStat'}),
    ('stat.ds-uart0.tx-buffers', {'id': 2, 'rpc': 'getStat'}),
    ('stat.ds-uart1.rx-buffers', {'id': 3, 'rpc': 'getStat'}),
    ('stat.ds-uart1.tx-buffers', {'id': 4, 'rpc': 'getStat'}),
    ('stat.ds-transparent.rx-buffers', {'id': 5, 'rpc': 'getStat'}),
    ('stat.ds-transparent.tx-buffers', {'id': 6, 'rpc': 'getStat'}),
    ('stat.packet-serial.rx-buffers', {'id': 7, 'rpc': 'getStat'}),
    ('stat.packet-serial.tx-buffers', {'id': 8, 'rpc': 'getStat'}),
    ('stat.radio.rx-buffers', {'id': 9, 'rpc': 'getStat'}),
    ('stat.radio.tx-buffers', {'id': 10, 'rpc': 'getStat'}),
    ('stat.radio.forwarded-unicasts', {'id': 11, 'rpc': 'getStat'}),
    ('stat.packet-serial.forwarded-unicasts', {'id': 12, 'rpc': 'getStat'}),
    ('stat.radio.forwarded-xcasts', {'id': 13, 'rpc': 'getStat'}),
    ('stat.packet-serial.forwarded-xcasts', {'id': 14, 'rpc': 'getStat'}),
])


@coroutine
@fix_up_target
def get_stats(gus, target, stat):
    """Get stats and info about the node.

    Args:
        gus (Gus): Grand Unified SNAP instance
        target (str, SNAP Address): Target Node
        stat (str): The name of the stat to query

    Returns:
        A dict with the requested stat(s)

    """
    data = {}

    try:
        val = None

        if stat in stat_lookup:
            stat_info = stat_lookup[stat]
            val = yield gus.callback_ucast_rpc(target, stat_info['rpc'], stat_info['id'])
            data[stat] = stat_info.get('formatter', formatters.format_noop)(first_or_none(val))
        elif stat == 'info':
            # Get all info stats
            for stat in stat_lookup:
                if stat.startswith('info'):
                    result = yield get_stats(gus, target, stat)
                    data.update(**result)
        elif stat == 'stat':
            # Get all stat stats
            for stat in stat_lookup:
                if stat.startswith('stat'):
                    result = yield get_stats(gus, target, stat)
                    data.update(**result)
        elif stat == 'all':
            # Get all info and stat stats
            for stat in stat_lookup:
                result = yield get_stats(gus, target, stat)
                data.update(**result)
        else:
            raise KeyError()
    except KeyError:
        raise ValueError('%s is not a valid stat name' % stat)

    raise Return(data)


@coroutine
@fix_up_target
def get_vmstat(gus, target, vmstat_code, vmstat_arg=None, vmstat_delay=None, retries=3, timeout=2):
    """Get the specified VMStat from the target.

    Equivalent of vmStat(code, arg).
    Does its own retries to ensure we get back a response to the vmstat code we originally sent.

    Args:
        gus (Gus): Grand Unified SNAP instance
        target (str, SNAP Address): Target Node
        vmstat_code (int): vmStat code
        vmstat_arg (int): vmStat argument

    Returns:
        A dict with the vmstat response

    """

    data = {}

    args = [_ for _ in [vmstat_code, vmstat_arg, vmstat_delay] if _ is not None]

    # TODO: Update this if we ever update the underlying api to tell the difference.
    for _ in xrange(retries):
        try:
            val = yield gus.callback_ucast_rpc(
                target,
                'vmStat',
                *args,
                callback_name='tellVmStat',
                retries=0,
                timeout=timeout
            )

        except NoResponseError:
            continue

        (word, d) = val

        word = word & 0xFFFF
        code = word & 0xFF
        hibyte = word >> 8

        if code != vmstat_code:
            # Not a response to our query!
            log.debug("Got tellVmStat response, but it wasn't for our query: %r, %r, %r", code, hibyte, d)
            continue

        data['code'] = code
        data['hibyte'] = hibyte
        data['data'] = d

        raise Return(data)

    raise NoResponseError()


@coroutine
@fix_up_target
def traceroute(gus, target):
    data = {}
    future = Future()

    def on_traceroute(node_addr, round_trip_time, hops):
        data['node'] = node_addr
        data['round_trip_time'] = round_trip_time
        data['hops'] = hops

        # Mark the traceroute operation as completed so that the coroutine can exit
        future.set_result(None)

    gus.on_traceroute(on_traceroute)

    gus.traceroute(target)

    # Wait for the future to resolve before exiting the coroutine
    yield future

    raise Return(data)


@coroutine
@fix_up_target
def scan_energy(gus, target, num=1, delay=0):
    """DEPRECATED - This uses a simple averaging of energy scans. Use percent_clear_energies() instead."""
    data = {}
    for i in xrange(num):
        val = yield gus.callback_ucast_rpc(target, 'scanEnergy')
        energies = first_or_none(val)

        # Unpack the response
        for i in xrange(len(energies)):
            data.setdefault('energy_channel_%02d' % i, []).append(ord(energies[i]))

        log.debug("DATA: %r", data)
        yield sleep(delay)
    else:
        for channel, scans in data.iteritems():
            data[channel] = float(sum(scans)) / len(scans)

    log.debug("DATA: %r", data)
    raise Return(data)


@coroutine
@fix_up_target
def raw_scan_energies(gus, target, num=1, delay=0):
    data = {}
    for i in xrange(num):
        val = yield gus.callback_ucast_rpc(target, 'scanEnergy')
        energies = first_or_none(val)

        # Unpack the response
        for i in xrange(len(energies)):
            data.setdefault(i, []).append(ord(energies[i]))

        log.debug("DATA: %r", data)
        yield sleep(delay)

    log.debug("DATA: %r", data)
    raise Return(data)


@coroutine
@fix_up_target
def percent_clear_energies(gus, target, num=1, delay=0, cutoff=90):

    raw_energies = yield raw_scan_energies(gus, target, num=num, delay=delay)

    raise Return(calculate_percent_clear(raw_energies, cutoff))


def calculate_percent_clear(energies, cutoff):
    """Calculates the weighted clear percentage of each channel relative to the least clear channel.

    The idea here is that each channel is assigned a "noisy" value that is affected by the cutoff value,
    where that value is the sum of weighted signal strength measurements. Higher-strength signal measurements are
    weighted more heavily, and the weight is determined by how close the measurement is to the cutoff value.

    The "percent clear" value for each channel represents how much less noisy that channel is compared to the weighted
    value of the noisiest channel.
    """

    weights = defaultdict(lambda: 0)
    for channel, scan_list in energies.iteritems():
        frequency_counts = Counter(scan_list)
        for scan_value in xrange(0, cutoff + 1):
            weights[channel] += frequency_counts[scan_value] * (cutoff + 1 - scan_value)
    max_weight = max(weights.values()) or 1

    def percent_of_max(weight):
        return 1 - (float(weight) / max_weight)

    percent_clear_dict = {channel: percent_of_max(weight) for (channel, weight) in weights.iteritems()}
    percent_clear_dict = {'clear_percent_channel_%02d' % channel: value for (channel, value) in percent_clear_dict.iteritems()}
    return percent_clear_dict


@coroutine
@fix_up_target
def start_output_interceptor(gus, target, intercept_stdout=True, intercept_stderr=True):
    data = {}
    if intercept_stdout:
        yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_STDIO)
    if intercept_stderr:
        yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_ERROR)
    yield gus.callback_ucast_rpc(target, 'ucastSerial', binascii.unhexlify(gus.address))
    raise Return(data)


@coroutine
@fix_up_target
def stop_output_interceptor(gus, target):
    data = {}
    yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_STDIO)
    yield gus.callback_ucast_rpc(target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_ERROR)
    raise Return(data)


def lookup_platform_by_id(module_id, platform_str=None):
    log.debug("Looking up module id %d, str %r", module_id, platform_str)

    candidates = []
    for k, v in SnapConstantsHelper.platforms_strings_to_ids.iteritems():
        if v == module_id:
            log.debug("Found a candidate: %s", k)
            candidates.append(k)
            # If it's an exact match, return that.
            if platform_str in SnapConstantsHelper.platforms_strings_to_ids and SnapConstantsHelper.platforms_strings_to_ids[platform_str] == module_id:
                return platform_str

    log.debug("Candidates: %r", candidates)

    if candidates:
        return candidates[0]
    else:
        return None


def lookup_platform_by_core_and_id(core, module_id, platform_str=None):
    log.debug("Looking up module part (to use with SNAPbuild) for core %s, module id %d, str %r", core, module_id, platform_str)

    # Extend the path so we can find firmware support packages
    ensure_path_contains_python_lib()

    try:
        # Import the firmware metadata based on the version
        firmware_pkg = "snap_firmware_" + core.replace('.', '_')
        core_module = importlib.import_module(firmware_pkg + ".metadata")
    except ImportError:
        core_module = None

    if core_module is None:
        log.error("Couldn't find firmware support package for core '%s', you may need to install it " +
                  "by running the following command:\n" +
                  "    pip install --extra-index-url https://update.synapse-wireless.com/pypi snap-firmware-%s",
                  core, core)
        return None

    # We rely on the snap-firmware-* metadata to figure out what a valid "module part" is now.
    # This sort of stuff *really* ought to be burned into the modules at MFG time,
    # but it isn't as of 2016-03-28.
    # So! What we're going to do here is a rough inverse of the lookup SNAPbuild will do when
    # it needs the part to get back to the CPU.
    # For C code support...we'll just have to hope that the antenna variants never have differing
    # linker offsets...

    # What we need is a "module_id" to "part number" mapping for the particular core.

    candidates = []
    supported_modules = core_module.get_supported_modules()

    all_metadata = {module: core_module.get_metadata(module, core, True) for module in supported_modules}

    log.debug("All Metadata: %r", all_metadata)

    for module, metadata in all_metadata.iteritems():
        if metadata['platform'] == module_id:
            candidates.append((module, core_module.get_module_platform_string(module), metadata))

    candidates = sorted(candidates)

    log.debug("Candidates: %r", candidates)

    if candidates:
        log.warning("Multiple parts (%s) match module ID %d", ",".join([c[0] for c in candidates]), module_id)
        log.info("Attempting to match by platform NV Param %s...", platform_str)
        for c in candidates:
            if c[1] == platform_str:
                log.info("Found a match by platform NV Param: %r", c)
                return c[0]

        log.warning("No match by platform NV Param %s found, falling back on first candidate: %r", platform_str, candidates[0])
        return candidates[0][0]
    else:
        return None


@coroutine
@fix_up_target
def get_node_core_module(gus, target, search_core=None):
    """Attempts to back out the Module Part info from vmStat(6).

    Figuring out the module part from the node itself is a bit of guesswork.
    The values returned here are sufficient for SNAPbuild, but may not be an exact match for your modules.
    If you *really* need an exact match, you should be using something like the platform NV param in your scripts.
    """

    data = {}

    result = yield get_vmstat(gus, target, 6)

    info = result['data'].split(',')
    if len(info) >= 8:
        core = info[0]
        module_id = int(info[5])
        platform_str = info[7]
        module = lookup_platform_by_core_and_id(search_core if search_core else core, module_id, platform_str=platform_str)

        data['core'] = core
        data['module'] = module
        data['platform'] = platform_str

    raise Return(data)


@coroutine
@fix_up_target
def ping(gus, target):
    """Pings a node and returns its link quality"""
    data = {}
    result = yield get_vmstat(gus, target, 5)
    data['lq'] = result['hibyte']
    raise Return(data)


@coroutine
@fix_up_target
def ota_script_upload(gus, target, filename=None, progress_cb=None, finished_cb=None, image=None):
    """Futures-compatible OTA Script upload"""

    data = {}

    if filename:
        # Read the image from the filename
        with open(filename, "rb") as f:
            image = ScriptsManager.getSnappyStringFromExport(f.read())
            if not image:
                raise Exception("Unable to load SNAPpy image from %s", filename)
    else:
        if not image:
            raise ValueError("Expected either filename or image as input, neither was provided")

    yield _ota_image_load(gus, target, image, progress_cb=progress_cb, finished_cb=finished_cb)

    data['spy'] = filename

    raise Return(data)


@coroutine
def _ota_image_load(gus, target, image, progress_cb=None, finished_cb=None):
    # Determine chunk size, allow for RPC_CRC and Packet CRC
    chunk_size = SnapConstantsHelper.getMaxPacketSize() - 30

    # Chunk the image
    def chunk(array, size):
        """Generator that yields sized chunks from array."""
        for i in xrange(0, len(array), size):
            yield array[i:i + size]

    # Send eraseImage
    yield gus.callback_ucast_rpc(target, 'eraseImage')

    image_size = len(image)

    chunks_done = []
    MAX_ATTEMPTS = 5

    # Send the chunks
    offset = 0
    for i, c in enumerate(chunk(image, chunk_size)):
        attempt = 0
        while True:
            attempt += 1
            log.debug("Attempt %d of %d writing chunk to %s at offset %d", attempt, MAX_ATTEMPTS, target, offset)
            if attempt > MAX_ATTEMPTS:
                raise Exception("Unable to load")

            # We'll do our own retries
            try:
                result = yield gus.callback_ucast_rpc(target, 'writeChunk', offset, c, callback_name='tellVmStat', timeout=2, retries=0)
                log.debug("Got result: %r", result)
            except NoResponseError:
                log.warn("Unable to load chunk %d on target %s", i, target)
                continue

            (word, d) = result

            word = word & 0xFFFF
            code = word & 0xFF

            if code != 2:
                log.warn("Unexpected tellVmStat from %s", target)
                continue

            written = d & 0xFFFF

            if written in chunks_done:
                log.warn("Got repeated notification for chunk: %d", written)
                continue

            if offset != written:
                log.error("Bad chunk write at %d, written %d", offset, written)
                continue

            log.debug("Successfully wrote chunk to %s at offset %d", target, offset)

            chunks_done.append(written)
            offset += len(c)

            if progress_cb:
                percent_complete = int(100 * offset / image_size)
                progress_cb(percent_complete)

            break

    # Send a reboot
    result = yield gus.callback_ucast_rpc(target, 'reboot')
    if result is None:
        raise Exception("Failed to reboot target %s", target)

    if finished_cb:
        finished_cb()


def _get_file_architecture(filename):
    """Return the (file_cpu, file_platform) for the given filename

    If the type cannot be determined then it will be set to None
    """
    file_cpu, file_platform = None, None
    if "MC1322x" in filename or "SM700" in filename:
        file_cpu = SnapConstantsHelper.CPU_ARM7
    elif "STM32W108" in filename:
        file_cpu = SnapConstantsHelper.CPU_ARM_CORTEX_M3
    # ATmega family
    elif "SM222" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM222_PFM
    elif "ATmega128RFR2" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega128RFR2_PFM
    elif "ATmega1284RFR2" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega1284RFR2_PFM
    elif "SM220" in filename or "RF220UF1" in filename or "SN220" in filename:
        # RF220UF1 is an SM220UF1 on an RF form factor carrier board
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM220UF1_PFM
    elif "RF220SU" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF220_SU
    elif "SM200" in filename or "RF200PU1" in filename or "RF200P81" in filename:
        # RF200PU1 is an SM200 module on an RF module form factor:
        # So is RF200P81
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.SM200_PFM
    elif "SS200" in filename or "RF200" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF200_PFM
    elif "RF266" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.RF266_PFM
    elif "ATmega128RFA1" in filename:
        file_cpu = SnapConstantsHelper.CPU_ATMEGA
        file_platform = SnapConstantsHelper.ATmega128RFA1

    return file_cpu, file_platform


def _get_layout(cpu):
    """We only really support the ATmega128RFA1 for OTA firmware upgrades"""
    from snaplayout import ATmega128RFA1_Layout
    layout_map = {SnapConstantsHelper.CPU_ATMEGA: ATmega128RFA1_Layout.ATmega128RFA1_Layout}

    return layout_map.get(cpu, None)


def _get_layout_from_filename(filename):
    file_cpu, file_platform = _get_file_architecture(filename)
    layout_class = _get_layout(file_cpu)

    log.info("Firmware file %s is for CPU %s, Platform %s, will use layout %s", filename, file_cpu, file_platform, layout_class)

    # Read the image from the filename
    with open(filename, "rb") as f:
        contents = f.read()

        # *.sfi files are compressed
        if os.path.splitext(filename)[1] == '.sfi':
            contents = bz2.decompress(contents)

    layout = layout_class(contents)

    if not layout:
        raise Exception("Unable to load Firmware image from %s" % filename)

    return layout


@coroutine
@fix_up_target
def ota_firmware_upload(gus, target, core, filename=None, progress_cb=None, finished_cb=None):
    """Futures-compatible OTA Firmware upload"""

    data = {}

    if filename is None:

        # Determine target's core/module type
        core_module = yield get_node_core_module(gus, target, search_core=core)

        module = core_module['module']

        data.update(**core_module)

        log.info("Target %s is a module %s, currently running core %s", target, module, data['core'])

        # Extend the path so we can find firmware support packages
        ensure_path_contains_python_lib()

        try:
            metadata_module_name = "snap_firmware_%s" % (core.replace('.', '_'),)
            metadata = importlib.import_module(metadata_module_name + '.metadata')
            log.debug("Looking for firmware for %s", metadata_module_name)

            filename = _get_file_path_for_version(metadata, module, core)
        except:
            log.exception("Couldn't find firmware module %s, you may need to `pip install --extra-index-url https://update.synapse-wireless.com/pypi %s`", metadata_module_name, metadata_module_name)
            raise
    else:
        # A specific filename was provided, try it.
        filename = os.path.abspath(os.path.join(os.path.curdir, os.path.expanduser(filename)))
        if not core:
            # Try to extract core from filename
            try:
                core = os.path.splitext(os.path.basename(filename))[0].split('V')[1]
                log.info("%s looks like it's version %s", filename, core)
            except:
                log.warning("Unable to determine core version from filename %s", filename)
                core = None

    log.info("Loading firmware %s onto module %s", filename, target)

    layout = _get_layout_from_filename(filename)

    # Get Bank Free Space
    result = yield gus.callback_ucast_rpc(target, 'getInfo', 16)
    if result is None:
        raise Exception("Couldn't get bank free space on %s" % target)

    free_space = result[0] & 0xFFFF

    if free_space == 0:
        raise Exception("Node %s reports it doesn't support OTA Firmware Upgrades" % target)

    image = layout.getContents(free_space)

    yield _ota_image_load(gus, target, image, progress_cb=progress_cb, finished_cb=finished_cb)

    log.warning('Pinging node %s to verify upgrade process is complete', target)

    yield sleep(3)
    try:
        response = yield get_vmstat(gus, target, 6, retries=15, timeout=10)
    except NoResponseError:
        raise Exception("Node %s never responded after upgrade" % target)

    info = response['data'].split(',')
    new_core = info[0]
    new_module_id = info[5]
    new_platform = info[7]
    if core:
        if new_core != core:
            raise Exception("Upgrading node %s did not complete successfully" % target)
    else:
        log.warning("Unable to determine core from filename, version loaded is %s", new_core)

    data['core'] = new_core
    data['platform'] = new_platform
    data['firmware'] = filename
    if 'module' not in data:
        data['module'] = new_module_id
    raise Return(data)
