import os
import logging
import re

log = logging.getLogger(__name__)

ADDR_RE = re.compile(r'^[0-9A-Fa-f]{6}$')


def _env_override():
    """Try to find SNAP_NODE_CACHE env var"""

    try:
        return os.environ['SNAP_NODE_CACHE']
    except:
        return None


def user_path():
    return _env_override() or os.path.expanduser(os.path.join('~', '.snap', 'nodes.cache'))


class NodeCache(object):
    """Simple cache of nodes"""

    def __init__(self):
        self.CACHE_FILE = user_path()
        self.cache = set()

    def _ensure_file_exists(self):
        with open(self.CACHE_FILE, 'a'):
            pass

    def load(self):
        self._ensure_file_exists()

        with open(self.CACHE_FILE, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if ADDR_RE.match(line):
                    self.cache.add(line)

        log.debug("Node Cache: %r", self.cache)

    def save(self):
        with open(self.CACHE_FILE, 'w') as f:
            for addr in self.cache:
                if ADDR_RE.match(addr):
                    f.write(addr)
                    f.write('\n')

    def __contains__(self, item):
        return item in self.cache

    def __len__(self):
        return len(self.cache)

    def __iter__(self):
        return self.cache.__iter__()

    def add(self, address):
        self.cache.add(address)

    def discard(self, address):
        self.cache.discard(address)

    def clear(self):
        self.cache.clear()

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()
