import logging

# Helper
from util import first_or_none

from tornado.gen import coroutine, Return

from snapconnect import snap

from node_cache import NodeCache

from .nvparam import check_and_set_nvparam

from .gus import NoResponseError

log = logging.getLogger(__name__)


CHANNELS = [4, 9, 7, 0, 10, 7, 12, 1, 2, 3, 5, 6, 8, 11, 13, 14, 15]


@coroutine
def ping_target(gus, target):
    try:
        val = yield gus.callback_ucast_rpc(target, 'getNetId')
    except NoResponseError:
        val = None
    raise Return(first_or_none(val))


@coroutine
def ping_on_channel(gus, target, channel):

    if target == gus.bridge_address:
        # it's the bridge node, see where its nvparams say it should be
        val = yield gus.callback_ucast_rpc(gus.bridge_address, 'loadNvParam', 4)
        channel = first_or_none(val)

        val = yield gus.callback_ucast_rpc(gus.bridge_address, 'loadNvParam', 3)
        net_id = first_or_none(val)

        raise Return((channel, net_id))

    else:
        # Move the bridge to the desired channel
        yield gus.callback_ucast_rpc(gus.bridge_address, 'setChannel', channel)
        yield gus.callback_ucast_rpc(gus.bridge_address, 'setNetId', 0xFFFF)

        val = yield ping_target(gus, target)
        net_id = first_or_none(val)

        raise Return((channel, net_id))


def move_channel_to_front(c):
    index = CHANNELS.index(c)

    if index >= 0:
        del CHANNELS[index]
        CHANNELS.insert(0, c)


@coroutine
def find_one_target(gus, target):

    for index, channel in enumerate(CHANNELS):
        loc = yield ping_on_channel(gus, target, channel)
        if loc[1]:
            log.info("Found target %s on channel %02d, net_id 0x%04x", target,
                     loc[0], loc[1])

            move_channel_to_front(channel)
            raise Return(loc)

    raise Exception("Target not found")


def get_targets_from_file(filename):
    targets = []

    for line in open(filename, 'rt'):
        line = line.strip()
        if line and len(line) > 0:
            if line[0] == '#':
                continue

            try:
                log.debug("Read %s from file", line)
                targets.append(line)
            except:
                log.exception("Couldn't parse address: %r", line)

    return targets


@coroutine
def move_target_to_network(gus, target, move_to_network):

    data = {}

    to = move_to_network.as_sc_params()
    channel = to.channel
    net_id = to.net_id
    enc_key = to.encryption_key
    enc_type = to.encryption_type

    # We just want to set/clear the CRC flags, and we can't assume
    # our target is running the nifty new 2.6.2 stuff
    val = yield gus.callback_ucast_rpc(target, 'loadNvParam', snap.NV_FEATURE_BITS_ID)
    current_features = first_or_none(val)
    network_features = to.feature_bits

    features = (current_features & 0xFAFF) | network_features

    yield check_and_set_nvparam(gus, target, snap.NV_CHANNEL_ID, channel)
    yield check_and_set_nvparam(gus, target, snap.NV_NETWORK_ID, net_id)
    yield check_and_set_nvparam(gus, target, snap.NV_AES128_KEY_ID, enc_key)
    yield check_and_set_nvparam(gus, target, snap.NV_AES128_ENABLE_ID, enc_type)
    yield check_and_set_nvparam(gus, target, snap.NV_FEATURE_BITS_ID, features)

    try:
        yield gus.callback_ucast_rpc(target, 'reboot')
    except NoResponseError:
        pass

    data.setdefault(target, dict())
    data[target]['to_channel'] = channel
    data[target]['to_net_id'] = "0x%04x" % net_id
    data[target]['to_enc_key'] = enc_key
    data[target]['to_enc_type'] = enc_type
    data[target]['to_features'] = "0x%04x" % features

    raise Return(data)


@coroutine
def finder(gus,
           targets,
           move_to_network=None):
    """Node finder

    Sweeps channels on the wildcard Network ID, pinging for target nodes.

    Args:
        gus (Gus): Grand Unified SNAP instance.
        targets (list[str]|str): Target address(es).
        move_to_network (str|None): Network to move nodes to once found.

    """

    data = {}

    log.debug("Node Finder, looking for %s, optionally moving to network %r",
              targets, move_to_network)

    if targets is None:
        targets = []
    else:
        if isinstance(targets, basestring):
            targets = [targets]
        targets = list(targets)

    with NodeCache() as cache:
        for t in targets:

            try:
                if t == "bridge":
                    t = gus.bridge_address

                cache.add(t)

                val = yield find_one_target(gus, t)

                target_data = data.get(t, {})

                target_data['address'] = t
                target_data['on_channel'] = val[0]
                target_data['on_net_id'] = "0x%04x" % (val[1] & 0xFFFF)

                data[t] = target_data

                if move_to_network:
                    try:
                        move_result = yield move_target_to_network(gus, t, move_to_network)
                        data[t].update(**move_result[t])
                    except:
                        log.exception("Couldn't move node %s to network %r", t,
                                      move_to_network)

            except:
                log.exception("Couldn't find node %s", t)
                data[t] = {}

    log.info("Rebooting bridge to restore its settings...")
    yield gus.callback_ucast_rpc(gus.bridge_address, 'reboot')

    raise Return(data)
