import logging
import platform
import os
import functools

from apy import EventScheduler

from serialwrapper.ComUtils import ComPortScanner, PROBE_VER_SNAP2, PROBE_VER_SNYF1
from serialwrapper.ComUtils import PROBE_RESULT_FOUND, PROBE_RESULT_READ_ERROR

from snap_profile.profile import Profile, SerialProfile

from tornado.ioloop import IOLoop, PeriodicCallback, TimeoutError
from tornado.concurrent import Future
from tornado.gen import coroutine, Return

log = logging.getLogger(__name__)


def get_com_port_scanner(scheduler):
    return ComPortScanner(scheduler)


@coroutine
def _probe(force=False, dry_run=False, sniffer=False, previously_scanned=None, progress_cb=None, finished_cb=None):

    node_kind = 'SNYF' if sniffer else 'SNAP'

    previously_scanned = previously_scanned or []

    result = Future()

    evQ = EventScheduler.EventScheduler()

    # HACK: workaround for absence of nonlocal in Py2 to avoid global scope
    class nonlocal:
        pass

    nonlocal.counter = 0

    def on_update(result):
        nonlocal.counter += 1
        if progress_cb:
            progress_cb(node_kind, nonlocal.counter)
        if result.result == PROBE_RESULT_FOUND or (force and result.result == PROBE_RESULT_READ_ERROR):
            if result.probeVer == PROBE_VER_SNAP2:
                log.info("Found SNAP node: %s", result)
            elif result.probeVer == PROBE_VER_SNYF1:
                log.info("Found SNYF node: %s", result)

    def on_complete(results):
        if finished_cb:
            finished_cb()
        log.debug("Final results (%d), %s", len(results), ', '.join(("%s" % r for r in results)))

        if not results:
            log.warn("No serial %s nodes found.", node_kind)
            result.set_result([])
            return

        known_devices = {}

        new_profiles = []

        for name in Profile.list():
            p = Profile.load(name)
            if p.kind == "serial":
                known_devices[p.device] = p

        for r in results:
            device = None
            if r.type == 2:
                # SS100
                device = "USB%s" % r.port
            elif r.type == 5:
                # SS200
                device = "SnapStick%s" % r.port
            else:
                # RS232
                if platform.system().lower().startswith('win'):
                    # COM###
                    device = os.path.basename(r.portStr).strip()
                else:
                    # /dev/*
                    device = r.port

            if device in known_devices:
                log.warn("Node %s already in profile: %s", device, known_devices[device].name)
                continue

            if r.result == PROBE_RESULT_FOUND or (force and r.result == PROBE_RESULT_READ_ERROR):
                log.info("Found a %s%s at %s", "Sniffer" if r.probeVer == PROBE_VER_SNYF1 else "SNAP node",
                         " with read error" if force and r.result == PROBE_RESULT_READ_ERROR else "", device)

                new_profile = SerialProfile()
                new_profile.name = os.path.basename(r.portStr).strip()
                new_profile.device = device
                if not dry_run:
                    new_profile.save()

                new_profiles.append(new_profile)

            else:
                log.warn("No %s node found at %s", node_kind, r)

        result.set_result(new_profiles)

    scanner = get_com_port_scanner(evQ)
    scanner.registerResultCallback(on_update)
    scanner.start(probeVer=(PROBE_VER_SNYF1 if sniffer else PROBE_VER_SNAP2), autoRunCallback=on_complete)

    PeriodicCallback(evQ.poll, 5).start()

    profiles = yield result
    raise Return(profiles)


def probe_serial_nodes(force=False, dry_run=False, timeout=180, progress_cb=None, finished_cb=None):
    snap_cb = functools.partial(_probe, force=force, sniffer=False, dry_run=dry_run, progress_cb=progress_cb, finished_cb=finished_cb)
    new_snap_profiles = []
    try:
        new_snap_profiles = IOLoop.current().run_sync(snap_cb, timeout=timeout)
    except TimeoutError:
        log.error("SNAP scan timed out or aborted!")

    log.debug("Got new SNAP profiles: %r", new_snap_profiles)

    snyf_cb = functools.partial(_probe, force=force, sniffer=True, dry_run=dry_run, progress_cb=progress_cb, finished_cb=finished_cb)
    new_snyf_profiles = []
    try:
        new_snyf_profiles = IOLoop.current().run_sync(snyf_cb, timeout=timeout)
    except TimeoutError:
        log.error("SNYF scan timed out or aborted!")

    log.debug("Got new Sniffer profiles: %r", new_snyf_profiles)

    return new_snap_profiles + new_snyf_profiles
