"""Utilities for gathering topologies from a mesh"""
from __future__ import print_function

import binascii
import logging

from collections import deque, Iterable, defaultdict

from tornado.gen import coroutine, Return

from .gus import NoResponseError

from .node import get_vmstat
from .node_cache import NodeCache
from .util import fix_up_target

log = logging.getLogger(__name__)


def _chunk_string(string, size):
    """Break a string into fixed-size chunks"""

    if not string:
        return []

    if not size:
        size = len(string)

    return [string[i:i + size] for i in range(0, len(string), size)]


def _parse_topo_response(src, response):
    """
    Parse the Topology (vmStat 12) response chunk.

    Returns a dict of Edges (or {} if the response could not be processed):
    * A, B: lq A->B (i.e. "how well B could hear A")
    * B, A: lq B->A (i.e. "how well A could hear B")
    """

    if not src:
        return {}

    if not response:
        return {}

    if len(response) != 5:
        return {}

    dst = binascii.hexlify(response[0:3])

    lq_src_to_dst = ord(response[3])
    lq_dst_to_src = ord(response[4])

    return {(src, dst): lq_src_to_dst,
            (dst, src): lq_dst_to_src}


def _mean(l):
    if isinstance(l, Iterable):
        return int(sum(l) / len(l))
    else:
        # If it's somehow not a list-ish thing, just return it.
        # We probably got called multiple times on the same topo.
        # (Mean of single value is that value.)
        return l


def _merge_topo_responses(responses):
    """Merge multiple node responses, averaging multiple lqs"""

    if not responses:
        return {}

    return {(src, dst): _mean(lqs) for (src, dst), lqs in responses.iteritems()}


@coroutine
@fix_up_target
def take_node_census(gus, target):
    """Take a census of a single node"""
    data = {}

    try:
        val = yield get_vmstat(gus, target, 12)
    except NoResponseError:
        raise Return({'root': target, 'topology': {}})

    parsed = (_parse_topo_response(target, s) for s in _chunk_string(val['data'], 5))

    with NodeCache() as cache:
        for response in parsed:
            for (src, dst), lq in response.iteritems():
                data[(src, dst)] = lq

                cache.add(src)
                cache.add(dst)

    raise Return({'root': target, 'topology': data})


@coroutine
@fix_up_target
def take_network_census(gus, target):
    """
    Take a census of an entire network, starting with a single target

    Breadth-first from the starting point.
    """
    data = defaultdict(list)

    nodes = deque()
    polled = set()
    nodes.append(target)

    while True:
        try:
            n = nodes.pop()
            while n in polled:
                n = nodes.pop()
        except IndexError:
            break

        try:
            log.debug("Getting Node Topology for: %r", n)
            polled.add(n)
            node_topo = yield take_node_census(gus, n)
            log.debug("Got Node Topology: %r", node_topo)

            for (src, dst), lq in node_topo['topology'].iteritems():
                data[(src, dst)].append(lq)

                if src not in polled:
                    nodes.append(src)
                if dst not in polled:
                    nodes.append(dst)

        except:
            log.exception("Gathering topology for node %s failed", n)
            continue

    merged = _merge_topo_responses(data)

    raise Return({'root': target, 'topology': merged})
